# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import EMEData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class EME(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_eme,
                ns=Namespace.EME,
                priority=40,
            )
        ]

    def _process_eme(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        encryption = stanza.getTag("encryption", namespace=Namespace.EME)
        if encryption is None:
            return

        name = encryption.getAttr("name")
        namespace = encryption.getAttr("namespace")
        if namespace is None:
            self._log.warning("No namespace on message")
            return

        properties.eme = EMEData(name=name, namespace=namespace)
        self._log.info("Found data: %s", properties.eme)



from test.lib.util import StanzaHandlerTest
from test.lib.util import utc_now

from nbxmpp import Message
from nbxmpp import Namespace
from nbxmpp.modules.retraction import Retraction
from nbxmpp.protocol import JID
from nbxmpp.structs import MAMData
from nbxmpp.structs import MessageProperties


class TestRetraction(StanzaHandlerTest):
    def test_parse_retract_direct_message(self):
        # language=XML
        xml = """
        <message type='chat' to='lord@capulet.example' id='retract-message-1'>
          <retract id="origin-id-1" xmlns='urn:xmpp:message-retract:1'/>
          <fallback xmlns="urn:xmpp:fallback:0"/>
          <body>This person attempted to retract a previous message</body>
          <store xmlns="urn:xmpp:hints"/>
        </message>
        """

        timestamp = utc_now()
        props = MessageProperties(
            JID.from_string("mockjid@jid.com"), timestamp=timestamp.timestamp()
        )
        message = Message(node=xml)
        Retraction._process_message_retraction(None, None, message, props)
        assert props.retraction is not None
        self.assertEqual(props.retraction.id, "origin-id-1")
        self.assertFalse(props.retraction.is_tombstone)
        self.assertEqual(props.retraction.timestamp, timestamp)

    def test_parse_retract_groupchat_message(self):
        # language=XML
        xml = """
        <message type='groupchat'
                 to='me@capulet.example/gajim'
                 from='room@component/retracter'>
          <retract id="stanza-id-1" xmlns='urn:xmpp:message-retract:1'/>
          <occupant-id xmlns="urn:xmpp:occupant-id:0" id="occupant-id" />
          <fallback xmlns="urn:xmpp:fallback:0"/>
          <body>This person attempted to retract a previous message</body>
          <store xmlns="urn:xmpp:hints"/>
        </message>
        """

        timestamp = utc_now()
        props = MessageProperties(
            JID.from_string("mockjid@jid.com"), timestamp=timestamp.timestamp()
        )
        message = Message(node=xml)
        Retraction._process_message_retraction(None, None, message, props)
        assert props.retraction is not None
        self.assertEqual(props.retraction.id, "stanza-id-1")
        self.assertFalse(props.retraction.is_tombstone)
        self.assertEqual(props.retraction.timestamp, timestamp)

    def test_parse_retract_tombstone_message(self):
        # language=XML
        xml = """
        <message type='groupchat'
                 to='me@capulet.example/gajim'
                 from='room@component/retracter'>
          <retracted id="ref-to-retract-message-id" xmlns='urn:xmpp:message-retract:1'/>
        </message>
        """

        timestamp = utc_now()
        mam = MAMData(
            id="1",
            query_id="query1",
            archive=JID.from_string("archive@jid.com"),
            namespace="mam:1",
            timestamp=timestamp.timestamp(),
        )
        props = MessageProperties(JID.from_string("mockjid@jid.com"), mam=mam)
        message = Message(node=xml)
        Retraction._process_message_retraction(None, None, message, props)
        assert props.retraction is not None
        self.assertEqual(props.retraction.id, "ref-to-retract-message-id")
        self.assertTrue(props.retraction.is_tombstone)
        self.assertEqual(props.retraction.timestamp, timestamp)

    def test_set_retraction(self):
        msg = Message()
        msg.setRetracted("some-complex-xmpp-message-id")
        self.assertEqual(
            msg.getTagAttr("retract", "id", Namespace.MESSAGE_RETRACT_1),
            "some-complex-xmpp-message-id",
        )



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging
import time

from nbxmpp.const import StreamState
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Error
from nbxmpp.protocol import Protocol
from nbxmpp.simplexml import Node
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import LogAdapter

if TYPE_CHECKING:
    from nbxmpp.client import Client

log = logging.getLogger("nbxmpp.smacks")


class Smacks:
    """
    This is Smacks is the Stream Management class. It takes care of requesting
    and sending acks. Also, it keeps track of the unhandled outgoing stanzas.

    The dispatcher has to be able to access this class to increment the
    number of handled stanzas
    """

    def __init__(self, client: Client) -> None:
        self._client = client
        self._out_h = 0  # Outgoing stanzas handled
        self._in_h = 0  # Incoming stanzas handled
        self._acked_h = 0  # Last acked stanza

        self._uqueue: list[Protocol] = []  # Unhandled stanzas queue
        self._old_uqueue: list[Protocol] = (
            []
        )  # Unhandled stanzas queue of the last session

        # Max number of stanzas in queue before making a request
        self.max_queue = 0

        self._sm_supported = False
        self.enabled = False  # If SM is enabled
        self._enable_sent = False  # If we sent 'enable'
        self.resumed = False  # If the session was resumed
        self.resume_in_progress = False
        self.resume_supported = False  # Does the session support resume

        self._session_id: str | None = None
        self._location: str | None = None

        self._log = LogAdapter(log, {"context": client.log_context})

        self.register_handlers()

    @property
    def sm_supported(self) -> bool:
        return self._sm_supported

    @sm_supported.setter
    def sm_supported(self, value: bool) -> None:
        self._log.info("Server supports detected: %s", value)
        self._sm_supported = value

    @property
    def resumeable(self) -> bool:
        return self._session_id is not None and self.resume_supported

    def delegate(self, stanza: Protocol) -> None:
        if stanza.getNamespace() != Namespace.STREAM_MGMT:
            return
        if stanza.getName() == "resumed":
            self._on_resumed(stanza)
        elif stanza.getName() == "failed":
            self._on_failed(None, stanza, None)

    def register_handlers(self) -> None:
        handlers = [
            StanzaHandler(
                name="enabled", callback=self._on_enabled, xmlns=Namespace.STREAM_MGMT
            ),
            StanzaHandler(
                name="failed", callback=self._on_failed, xmlns=Namespace.STREAM_MGMT
            ),
            StanzaHandler(
                name="r", callback=self._send_ack, xmlns=Namespace.STREAM_MGMT
            ),
            StanzaHandler(name="a", callback=self._on_ack, xmlns=Namespace.STREAM_MGMT),
        ]

        for handler in handlers:
            self._client.register_handler(handler)

    def send_enable(self) -> None:
        if not self.sm_supported:
            return

        if self._client.sm_disabled:
            return

        enable = Node(Namespace.STREAM_MGMT + " enable", attrs={"resume": "true"})
        self._client.send_nonza(enable, now=False)
        self._log.debug("Send enable")
        self._enable_sent = True

    def _on_enabled(self, _con: Any, stanza: Protocol, _properties: Any) -> None:
        if self.enabled:
            self._log.error('Received "enabled", but SM is already enabled')
            return
        resume = stanza.getAttr("resume")
        if resume in ("true", "True", "1"):
            self.resume_supported = True
            self._session_id = stanza.getAttr("id")

        self._location = stanza.getAttr("location")
        self.enabled = True
        self._log.info(
            "Received enabled, location: %s, resume supported: %s, " "session-id: %s",
            self._location,
            resume,
            self._session_id,
        )

    def count_incoming(self, name: str) -> None:
        if not self.enabled:
            # Dont count while we didnt receive 'enabled'
            return
        if name in ("a", "r", "resumed", "enabled"):
            return
        self._log.debug("IN, %s", name)
        self._in_h += 1

    def save_in_queue(self, stanza: Protocol) -> None:
        if not self._enable_sent and not self.resumed:
            # We did not yet sent 'enable' so the server
            # will not count our stanzas
            return

        # Make a full copy so we dont run into problems when
        # the stanza is modified after sending for some reason
        # TODO: Make also copies of Protocol.Error objects
        if not isinstance(stanza, Error):
            stanza = type(stanza)(node=str(stanza))

        self._add_delay(stanza)

        self._uqueue.append(stanza)
        self._log.debug("OUT, %s", stanza.getName())
        self._out_h += 1

        if len(self._uqueue) > self.max_queue:
            self._request_ack()
        # Send an ack after 100 unacked messages
        if (self._in_h - self._acked_h) > 100:
            self._send_ack()

    def _add_delay(self, stanza: Protocol) -> None:
        if stanza.getName() != "message":
            return

        if stanza.getType() not in ("chat", "groupchat"):
            return

        timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        attrs = {"stamp": timestamp}

        if stanza.getType() != "groupchat":
            # Dont leak our JID to Groupchats
            attrs["from"] = str(self._client.get_bound_jid())

        stanza.addChild("delay", namespace=Namespace.DELAY2, attrs=attrs)

    def _resend_queue(self) -> None:
        """
        Resends unsent stanzas when a new session is established.
        This way there won't be any lost outgoing messages even on failed
        smacks resumes (but message duplicates are possible).
        """
        if not self._old_uqueue:
            return
        self._log.info("Resend %s stanzas", len(self._old_uqueue))
        for stanza in self._old_uqueue:
            # Use dispatcher so we increment the counter
            self._client.send_stanza(stanza)
        self._old_uqueue = []

    def resume_request(self) -> None:
        if self._session_id is None:
            self._log.error("Attempted to resume without a valid session id")
            return

        # Save old messages in an extra "queue" to avoid race conditions
        # and to make it possible to replay stanzas even when resuming fails
        # Add messages here (instead of overwriting) so that repeated
        # connection errors don't delete unacked stanzas
        # (uqueue should be empty in this case anyways)
        self._old_uqueue += self._uqueue
        self._uqueue = []

        resume = Node(
            Namespace.STREAM_MGMT + " resume",
            attrs={"h": self._in_h, "previd": self._session_id},
        )

        self._acked_h = self._in_h
        self.resume_in_progress = True
        self._client.send_nonza(resume, now=False)

    def _on_resumed(self, stanza: Protocol) -> None:
        """
        Checks if the number of stanzas sent are the same as the
        number of stanzas received by the server. Resends stanzas not received
        by the server in the last session.
        """
        self._log.info("Session resumption succeeded, session-id: %s", self._session_id)
        self._validate_ack(stanza, self._old_uqueue)
        # Set our out h to the h we received
        self._out_h = int(stanza.getAttr("h"))
        self.enabled = True
        self.resumed = True
        self.resume_in_progress = False
        self._client.set_state(StreamState.RESUME_SUCCESSFUL)
        self._resend_queue()

    def _send_ack(self, *args: Any) -> None:
        ack = Node(Namespace.STREAM_MGMT + " a", attrs={"h": self._in_h})
        self._acked_h = self._in_h
        self._log.debug("Send ack, h: %s", self._in_h)
        self._client.send_nonza(ack, now=False)

    def close_session(self) -> None:
        # We end the connection deliberately
        # Reset the state -> no resume
        self._log.info("Close session")
        self._reset_state()

    def _request_ack(self) -> None:
        request = Node(Namespace.STREAM_MGMT + " r")
        self._log.debug("Request ack")
        self._client.send_nonza(request, now=False)

    def _on_ack(self, _stream: Any, stanza: Protocol, _properties: Any) -> None:
        if not self.enabled:
            return
        self._log.debug("Ack received, h: %s", stanza.getAttr("h"))
        self._validate_ack(stanza, self._uqueue)

    def _validate_ack(self, stanza: Protocol, queue: list[Protocol]) -> None:
        """
        Checks if the number of stanzas sent are the same as the
        number of stanzas received by the server. Pops stanzas that were
        handled by the server from the queue.
        """
        count_server = stanza.getAttr("h")
        if count_server is None:
            self._log.error("Server did not send h attribute")
            return

        count_server = int(count_server)
        diff = self._out_h - count_server
        queue_size = len(queue)
        if diff < 0:
            self._log.error(
                "Mismatch detected, our h: %d, server h: %d, queue: %d",
                self._out_h,
                count_server,
                queue_size,
            )
            # Don't accumulate all messages in this case
            # (they would otherwise all be resent on the next reconnect)
            queue = []

        elif queue_size < diff:
            self._log.error(
                "Mismatch detected, our h: %d, server h: %d, queue: %d",
                self._out_h,
                count_server,
                queue_size,
            )
        else:
            self._log.debug(
                "Validate ack, our h: %d, server h: %d, queue: %d",
                self._out_h,
                count_server,
                queue_size,
            )
            self._log.debug("Removing %d stanzas from queue", queue_size - diff)

            while len(queue) > diff:
                queue.pop(0)

    def _on_failed(self, _stream: Any, stanza: Protocol, _properties: Any) -> None:
        """
        This can be called after 'enable' and 'resume'
        """

        self._log.info("Negotiation failed")
        error_text = stanza.getTagData("text")
        if error_text is not None:
            self._log.info(error_text)

        if stanza.getTag("item-not-found") is not None:
            self._log.info("Session timed out, last server h: %s", stanza.getAttr("h"))
            self._validate_ack(stanza, self._old_uqueue)
        else:
            for tag in stanza.getChildren():
                if tag.getName() != "text":
                    self._log.info(tag.getName())

        if self.resume_in_progress:
            # Reset state before sending Bind, because otherwise stanza
            # will be counted and ack will be requested.
            # _reset_state() also resets resume_in_progress
            self._reset_state()
            self._client.set_state(StreamState.RESUME_FAILED)

        self._reset_state()

    def _reset_state(self) -> None:
        # Reset all values to default
        self._out_h = 0
        self._in_h = 0
        self._acked_h = 0

        self._uqueue = []
        self._old_uqueue = []

        self.enabled = False
        self._enable_sent = False
        self.resumed = False
        self.resume_in_progress = False
        self.resume_supported = False

        self._session_id = None
        self._location = None
        self._enable_sent = False



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from collections.abc import Generator

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.simplexml import Node
from nbxmpp.structs import BlockingProperties
from nbxmpp.structs import BlockingPush
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.types import BlockingReportValues

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Blocking(BaseModule):
    def __init__(self, client: Client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                priority=15,
                callback=self._process_blocking_push,
                typ="set",
                ns=Namespace.BLOCKING,
            ),
        ]

    @iq_request_task
    def request_blocking_list(self) -> Generator[set[JID] | Iq | None, Iq]:
        _task = yield

        result = yield _make_blocking_list_request()
        if result.isError():
            raise StanzaError(result)

        blocklist = result.getTag("blocklist", namespace=Namespace.BLOCKING)
        if blocklist is None:
            raise MalformedStanzaError("blocklist node missing", result)

        blocked: set[JID] = set()
        for item in blocklist.getTags("item"):
            try:
                jid = JID.from_string(item.getAttr("jid"))
            except Exception:
                self._log.info("Invalid JID: %s", item.getAttr("jid"))
                continue

            blocked.add(jid)

        self._log.info("Received blocking list: %s", blocked)
        yield blocked

    @iq_request_task
    def block(self, jids: list[JID], report: BlockingReportValues | None = None):

        _task = yield

        response = yield _make_block_request(jids, report)
        yield process_response(response)

    @iq_request_task
    def unblock(self, jids: list[JID]):
        _task = yield

        response = yield _make_unblock_request(jids)
        yield process_response(response)

    @staticmethod
    def _process_blocking_push(
        client: Client, stanza: Iq, properties: BlockingProperties
    ) -> None:

        unblock = stanza.getTag("unblock", namespace=Namespace.BLOCKING)
        if unblock is not None:
            properties.blocking = _parse_push(unblock)

        block = stanza.getTag("block", namespace=Namespace.BLOCKING)
        if block is not None:
            properties.blocking = _parse_push(block)

        reply = stanza.buildSimpleReply("result")
        client.send_stanza(reply)


def _make_blocking_list_request() -> Iq:
    iq = Iq("get", Namespace.BLOCKING)
    iq.setQuery("blocklist")
    return iq


def _make_block_request(jids: list[JID], report: BlockingReportValues | None) -> Iq:

    iq = Iq("set", Namespace.BLOCKING)
    query = iq.setQuery(name="block")
    for jid in jids:
        item = query.addChild(name="item", attrs={"jid": str(jid)})
        if report in ("spam", "abuse"):
            action = item.addChild(name="report", namespace=Namespace.REPORTING)
            action.setTag(report)
    return iq


def _make_unblock_request(jids: list[JID]) -> Iq:
    iq = Iq("set", Namespace.BLOCKING)
    query = iq.setQuery(name="unblock")
    for jid in jids:
        query.addChild(name="item", attrs={"jid": str(jid)})
    return iq


def _parse_push(node: Node) -> BlockingPush:
    items = node.getTags("item")
    if not items:
        return BlockingPush(block=set(), unblock=set(), unblock_all=True)

    jids: set[JID] = set()
    for item in items:
        jid = item.getAttr("jid")
        if not jid:
            continue

        try:
            jid = JID.from_string(jid)
        except Exception:
            continue

        jids.add(jid)

    block: set[JID] = set()
    unblock: set[JID] = set()
    if node.getName() == "block":
        block = jids
    else:
        unblock = jids

    return BlockingPush(block=block, unblock=unblock, unblock_all=False)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import AvatarState
from nbxmpp.const import PresenceType
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Presence
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class VCardAvatar(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._process_avatar,
                ns=Namespace.VCARD_UPDATE,
                priority=15,
            )
        ]

    def _process_avatar(
        self, _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        if properties.type != PresenceType.AVAILABLE:
            return

        update = stanza.getTag("x", namespace=Namespace.VCARD_UPDATE)
        if update is None:
            return

        avatar_sha = update.getTagData("photo")
        if avatar_sha is None:
            properties.avatar_state = AvatarState.NOT_READY
            self._log.info("%s is not ready to promote an avatar", stanza.getFrom())
            # Empty update element, ignore
            return

        if avatar_sha == "":
            properties.avatar_state = AvatarState.EMPTY
            self._log.info("%s empty avatar advertised", stanza.getFrom())
            return

        properties.avatar_sha = avatar_sha
        properties.avatar_state = AvatarState.ADVERTISED
        self._log.info("%s advertises %s", stanza.getFrom(), avatar_sha)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import CorrectionData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Correction(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_correction,
                ns=Namespace.CORRECT,
                priority=15,
            ),
        ]

    def _process_message_correction(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        replace = stanza.getTag("replace", namespace=Namespace.CORRECT)
        if replace is None:
            return

        id_ = replace.getAttr("id")
        if id_ is None:
            self._log.warning("Correcton without id attribute")
            self._log.warning(stanza)
            return

        if stanza.getID() == id_:
            self._log.warning("correcton id == message id")
            self._log.warning(stanza)
            return

        properties.correction = CorrectionData(id_)



# Copyright (C) 2021 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from collections.abc import Callable

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_FORBIDDEN
from nbxmpp.protocol import ERR_SERVICE_UNAVAILABLE
from nbxmpp.protocol import Error
from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IqProperties
from nbxmpp.structs import LastActivityData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class LastActivity(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._answer_request,
                priority=60,
                typ="get",
                ns=Namespace.LAST,
            ),
        ]

        self._idle_func: Callable[..., int] | None = None
        self._allow_reply_func: Callable[..., bool] | None = None

    def disable(self) -> None:
        self._idle_func = None

    def set_idle_func(self, func: Callable[..., int]) -> None:
        self._idle_func = func

    def set_allow_reply_func(self, func: Callable[..., bool]) -> None:
        self._allow_reply_func = func

    @iq_request_task
    def request_last_activity(self, jid: str):
        _task = yield

        response = yield _make_request(jid)
        if response.isError():
            raise StanzaError(response)

        yield _parse_response(response)

    def _answer_request(
        self, _client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        self._log.info("Request received from %s", stanza.getFrom())
        if self._idle_func is None:
            self._client.send_stanza(Error(stanza, ERR_SERVICE_UNAVAILABLE))
            raise NodeProcessed

        if self._allow_reply_func is not None:
            if not self._allow_reply_func(stanza.getFrom()):
                self._client.send_stanza(Error(stanza, ERR_FORBIDDEN))
                raise NodeProcessed

        seconds = self._idle_func()
        iq = stanza.buildReply("result")
        query = iq.getQuery()
        query.setAttr("seconds", seconds)
        self._log.info("Send last activity: %s", seconds)
        self._client.send_stanza(iq)
        raise NodeProcessed


def _make_request(jid: str) -> Iq:
    return Iq("get", queryNS=Namespace.LAST, to=jid)


def _parse_response(response: Iq) -> LastActivityData:
    query = response.getQuery()
    seconds = query.getAttr("seconds")

    try:
        seconds = int(seconds)
    except Exception:
        raise MalformedStanzaError("seconds attribute invalid", response)

    return LastActivityData(seconds=seconds, status=query.getData())



import unittest

from nbxmpp.stringprep import check_bidi
from nbxmpp.stringprep import nameprep
from nbxmpp.stringprep import nodeprep
from nbxmpp.stringprep import resourceprep
from nbxmpp.stringprep import saslprep


class TestBidi(unittest.TestCase):
    def test_empty_string(self):
        check_bidi("")

    def test_L_RAL_violation(self):
        with self.assertRaises(ValueError):
            check_bidi("\u05be\u0041")


class TestNodeprep(unittest.TestCase):
    def test_map_to_nothing(self):
        self.assertEqual(
            "ix",
            nodeprep("I\u00ADX"),
            "Nodeprep requirement: map SOFT HYPHEN to nothing",
        )

    def test_case_fold(self):
        self.assertEqual(
            "ssa", nodeprep("ßA"), "Nodeprep requirement: map ß to ss, A to a"
        )

    def test_nfkc(self):
        self.assertEqual("a", nodeprep("\u00AA"), "Nodeprep requirement: NFKC")
        self.assertEqual("ix", nodeprep("\u2168"), "Nodeprep requirement: NFKC")

    def test_prohibited_character(self):
        with self.assertRaisesRegex(
            ValueError,
            r"U\+0007",
            msg="Nodeprep requirement: prohibited character (C.2.1)",
        ):
            nodeprep("\u0007")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+200e",
            msg="Nodeprep requirement: prohibited character (C.8)",
        ):
            nodeprep("\u200E")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+003e",
            msg="Nodeprep requirement: prohibited character (custom)",
        ):
            nodeprep(">")

    def test_unassigned(self):
        with self.assertRaises(ValueError, msg="Nodeprep requirement: unassigned"):
            nodeprep("\u0221", allow_unassigned=False)

        with self.assertRaises(ValueError, msg="enforce no unassigned by default"):
            nodeprep("\u0221")

        self.assertEqual("\u0221", nodeprep("\u0221", allow_unassigned=True))


class TestNameprep(unittest.TestCase):
    def test_map_to_nothing(self):
        self.assertEqual(
            "ix",
            nameprep("I\u00ADX"),
            "Nameprep requirement: map SOFT HYPHEN to nothing",
        )

    def test_case_fold(self):
        self.assertEqual(
            "ssa", nameprep("ßA"), "Nameprep requirement: map ß to ss, A to a"
        )

    def test_nfkc(self):
        self.assertEqual("a", nodeprep("\u00AA"), "Nameprep requirement: NFKC")
        self.assertEqual("ix", nodeprep("\u2168"), "Nameprep requirement: NFKC")

    def test_prohibited_character(self):
        with self.assertRaisesRegex(
            ValueError,
            r"U\+06dd",
            msg="Nameprep requirement: prohibited character (C.2.2)",
        ):
            nameprep("\u06DD")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+e000",
            msg="Nameprep requirement: prohibited character (C.3)",
        ):
            nameprep("\uE000")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+1fffe",
            msg="Nameprep requirement: prohibited character (C.4)",
        ):
            nameprep("\U0001FFFE")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+d800",
            msg="Nameprep requirement: prohibited character (C.5)",
        ):
            nameprep("\uD800")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+fff9",
            msg="Nameprep requirement: prohibited character (C.6)",
        ):
            nameprep("\uFFF9")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+2ff0",
            msg="Nameprep requirement: prohibited character (C.7)",
        ):
            nameprep("\u2FF0")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+e0001",
            msg="Nameprep requirement: prohibited character (C.9)",
        ):
            nameprep("\U000E0001")

    def test_unassigned(self):
        with self.assertRaises(ValueError, msg="Nameprep requirement: unassigned"):
            nameprep("\u0221", allow_unassigned=False)

        with self.assertRaises(ValueError, msg="enforce no unassigned by default"):
            nameprep("\u0221")

        self.assertEqual("\u0221", nameprep("\u0221", allow_unassigned=True))


class TestResourceprep(unittest.TestCase):
    def test_map_to_nothing(self):
        self.assertEqual(
            "IX",
            resourceprep("I\u00ADX"),
            "Resourceprep requirement: map SOFT HYPHEN to nothing",
        )

    def test_nfkc(self):
        self.assertEqual("a", resourceprep("\u00AA"), "Resourceprep requirement: NFKC")
        self.assertEqual("IX", resourceprep("\u2168"), "Resourceprep requirement: NFKC")

    def test_prohibited_character(self):
        with self.assertRaisesRegex(
            ValueError,
            r"U\+0007",
            msg="Resourceprep requirement: " "prohibited character (C.2.1)",
        ):
            resourceprep("\u0007")

        with self.assertRaisesRegex(
            ValueError,
            r"U\+200e",
            msg="Resourceprep requirement: " "prohibited character (C.8)",
        ):
            resourceprep("\u200E")

    def test_unassigned(self):
        with self.assertRaises(ValueError, msg="Resourceprep requirement: unassigned"):
            resourceprep("\u0221", allow_unassigned=False)

        with self.assertRaises(ValueError, msg="enforce no unassigned by default"):
            resourceprep("\u0221")

        self.assertEqual("\u0221", resourceprep("\u0221", allow_unassigned=True))


class TestSASLprep(unittest.TestCase):
    def test_map_to_nothing(self):
        self.assertEqual(
            "IX",
            saslprep("I\u00ADX"),
            "SASLprep requirement: map SOFT HYPHEN to nothing",
        )

    def test_nfkc(self):
        self.assertEqual("a", saslprep("\u00AA"), "SASLprep requirement: NFKC")
        self.assertEqual("IX", saslprep("\u2168"), "SASLprep requirement: NFKC")

    def test_case_fold(self):
        self.assertNotEqual(
            "user", saslprep("USER"), "SASLprep requirement: No CaseFold"
        )

    def test_prohibited_character(self):
        with self.assertRaisesRegex(
            ValueError,
            r"U\+0007",
            msg="SASLprep requirement: " "prohibited character (C.2.1)",
        ):
            saslprep("\u0007")

    def test_bidi_check(self):
        with self.assertRaises(ValueError, msg="SASLprep requirement: bidi check"):
            saslprep("\u0627\u0031")

    def test_unassigned(self):
        with self.assertRaises(ValueError, msg="SASLprep requirement: unassigned"):
            saslprep("\u0221", allow_unassigned=False)

        with self.assertRaises(ValueError, msg="enforce no unassigned by default"):
            saslprep("\u0221")

        self.assertEqual("\u0221", saslprep("\u0221", allow_unassigned=True))



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime as dt
import logging
import re
import time

log = logging.getLogger("nbxmpp.m.date_and_time")

PATTERN_DATETIME = re.compile(
    r"""
    ([0-9]{4}-[0-9]{2}-[0-9]{2})    # Date
    T                               # Separator
    ([0-9]{2}:[0-9]{2}:[0-9]{2})    # Time
    (?P<frac>\.[0-9]{0,6})?         # Fractual Seconds
    [0-9]*                          # lose everything > 6
    (Z|[-+][0-9]{2}:[0-9]{2})       # UTC Offset
    $                               # End of String
""",
    re.VERBOSE,
)

ZERO = dt.timedelta(0)
HOUR = dt.timedelta(hours=1)
SECOND = dt.timedelta(seconds=1)

STDOFFSET = dt.timedelta(seconds=-time.timezone)
if time.daylight:
    dstoffset = dt.timedelta(seconds=-time.altzone)
else:
    dstoffset = STDOFFSET

DSTOFFSET = dstoffset
DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(dt.tzinfo):
    """
    A class capturing the platform's idea of local time.
    May result in wrong values on historical times in
    timezones where UTC offset and/or the DST rules had
    changed in the past.
    """

    def fromutc(self, dt_input: dt.datetime) -> dt.datetime:
        stamp = (dt_input - dt.datetime(1970, 1, 1, tzinfo=self)) // SECOND
        args = time.localtime(stamp)[:6]
        dst_diff = DSTDIFF // SECOND
        # Detect fold
        fold = args == time.localtime(stamp - dst_diff)
        return dt.datetime(
            *args, microsecond=dt_input.microsecond, tzinfo=self, fold=fold
        )

    def utcoffset(self, dt_input: dt.datetime) -> dt.timedelta:  # type: ignore
        if self._isdst(dt_input):
            return DSTOFFSET
        return STDOFFSET

    def dst(self, dt_input: dt.datetime) -> dt.timedelta:  # type: ignore
        if self._isdst(dt_input):
            return DSTDIFF
        return ZERO

    def tzname(self, _dt_input: dt.datetime) -> str:  # type: ignore
        return "local"

    @staticmethod
    def _isdst(dt_input: dt.datetime) -> bool:
        tt = (
            dt_input.year,
            dt_input.month,
            dt_input.day,
            dt_input.hour,
            dt_input.minute,
            dt_input.second,
            dt_input.weekday(),
            0,
            0,
        )
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0


def create_tzinfo(
    hours: int = 0, minutes: int = 0, tz_string: str | None = None
) -> dt.timezone | None:
    if tz_string is None:
        return dt.timezone(dt.timedelta(hours=hours, minutes=minutes))

    if tz_string.lower() == "z":
        return dt.timezone.utc

    try:
        hours, minutes = map(int, tz_string.split(":"))
    except Exception:
        log.warning("Wrong tz string: %s", tz_string)
        return None

    if hours not in range(-24, 24):
        log.warning("Wrong tz string: %s", tz_string)
        return None

    if minutes not in range(59):
        log.warning("Wrong tz string: %s", tz_string)
        return None

    if hours in (24, -24) and minutes != 0:
        log.warning("Wrong tz string: %s", tz_string)
        return None
    return dt.timezone(dt.timedelta(hours=hours, minutes=minutes))


def parse_datetime(
    timestring: str | None,
    check_utc: bool = False,
    convert: str | None = "utc",
    epoch: bool = False,
) -> dt.datetime | float | None:
    """
    Parse a XEP-0082 DateTime Profile String

    :param timestring: a XEP-0082 DateTime profile formated string

    :param check_utc:  if True, returns None if timestring is not
                       a timestring expressing UTC

    :param convert:    convert the given timestring to utc or local time

    :param epoch:      if True, returns the time in epoch

    Examples:
    '2017-11-05T01:41:20Z'
    '2017-11-05T01:41:20.123Z'
    '2017-11-05T01:41:20.123+05:00'

    return a datetime or epoch
    """
    if timestring is None:
        return None
    if convert not in (None, "utc", "local"):
        raise TypeError('"%s" is not a valid value for convert')

    match = PATTERN_DATETIME.match(timestring)
    if match is None:
        return None

    timestring = "".join(match.groups(""))
    strformat = "%Y-%m-%d%H:%M:%S%z"
    if match.group("frac"):
        # Fractional second addendum to Time
        strformat = "%Y-%m-%d%H:%M:%S.%f%z"

    try:
        date_time = dt.datetime.strptime(timestring, strformat)
    except ValueError:
        return None

    if not 1 < date_time.year < 9999:
        # Raise/Reduce MIN/MAX year so converting to different
        # timezones cannot get out of range
        return None

    if check_utc:
        if convert != "utc":
            raise ValueError('check_utc can only be used with convert="utc"')

        if date_time.tzinfo != dt.timezone.utc:
            return None

        if epoch:
            return date_time.timestamp()
        return date_time

    if convert == "utc":
        date_time = date_time.astimezone(dt.timezone.utc)
        if epoch:
            return date_time.timestamp()
        return date_time

    if epoch:
        # epoch is always UTC, use convert='utc' or check_utc=True
        raise ValueError("epoch not available while converting to local")

    if convert == "local":
        date_time = date_time.astimezone(LocalTimezone())
        return date_time

    # convert=None
    return date_time


def get_local_time() -> tuple[str, str]:
    formated_time = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
    isdst = time.localtime().tm_isdst
    zone = -(time.timezone, time.altzone)[isdst] / 60.0
    zone = (zone / 60, abs(zone % 60))
    tzo = "%+03d:%02d" % zone
    return formated_time, tzo



from __future__ import annotations

from typing import Any
from typing import cast

import functools
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence

# Copied from https://codeberg.org/jssfr/aioxmpp/src/branch/devel/aioxmpp/structs.py


@functools.total_ordering
class LanguageTag:
    """
    Implementation of a language tag. This may be a fully RFC5646 compliant
    implementation some day, but for now it is only very simplistic stub.

    There is no input validation of any kind.
    """

    __slots__ = ("_tag",)

    def __init__(self, *, tag: str):
        if not tag:
            raise ValueError("tag cannot be empty")

        self._tag = tag

    @property
    def match_str(self) -> str:
        """
        The string which is used for matching two language tags. This is the
        lower-cased version of `print_str`.
        """
        return self._tag.lower()

    @property
    def print_str(self) -> str:
        """
        The stringified language tag.
        """
        return self._tag

    @classmethod
    def fromstr(cls, s: str) -> LanguageTag:
        """
        Create a language tag from the given string `s`.
        """
        return cls(tag=s)

    def __str__(self) -> str:
        return self.print_str

    def __eq__(self, other: Any) -> bool:
        try:
            return self.match_str == other.match_str
        except AttributeError:
            return False

    def __lt__(self, other: Any) -> bool:
        try:
            return self.match_str < other.match_str
        except AttributeError:
            return NotImplemented

    def __le__(self, other: Any) -> bool:
        try:
            return self.match_str <= other.match_str
        except AttributeError:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.match_str)

    def __repr__(self) -> str:
        return "<{}.{}.fromstr({!r})>".format(
            type(self).__module__, type(self).__qualname__, str(self)
        )


class LanguageRange:
    """
    Implementation of a language range. This may be a fully RFC4647 compliant
    implementation some day, but for now it is only very simplistic stub.

    There is no input validation of any kind.

    LanguageRange instances compare and hash case-insensitively.
    """

    __slots__ = ("_tag",)

    WILDCARD: LanguageRange

    def __init__(self, *, tag: str):
        if not tag:
            raise ValueError("range cannot be empty")

        self._tag = tag

    @property
    def match_str(self) -> str:
        """
        The string which is used for matching two language tags. This is the
        lower-cased version of the `print_str`.
        """
        return self._tag.lower()

    @property
    def print_str(self) -> str:
        """
        The stringified language tag.
        """
        return self._tag

    @classmethod
    def fromstr(cls, s: str) -> LanguageRange:
        """
        Create a language tag from the given string `s`.
        """
        if s == "*":
            return cls.WILDCARD
        return cls(tag=s)

    def __str__(self) -> str:
        return self.print_str

    def __eq__(self, other: Any) -> bool:
        try:
            return self.match_str == other.match_str
        except AttributeError:
            return False

    def __hash__(self) -> int:
        return hash(self.match_str)

    def __repr__(self) -> str:
        return "<{}.{}.fromstr({!r})>".format(
            type(self).__module__, type(self).__qualname__, str(self)
        )

    def strip_rightmost(self) -> LanguageRange:
        """
        Strip the rightmost part of the language range. If the new rightmost
        part is a singleton or ``x`` (i.e. starts an extension or private use
        part), it is also stripped.

        Return the newly created `LanguageRange`.
        """

        parts = self.print_str.split("-")
        parts.pop()
        if parts and len(parts[-1]) == 1:
            parts.pop()
        return type(self).fromstr("-".join(parts))


LanguageRange.WILDCARD = LanguageRange(tag="*")


def basic_filter_languages(
    languages: Sequence[LanguageTag], ranges: Iterable[LanguageRange]
) -> Iterator[LanguageTag]:
    """
    Filter languages using the string-based basic filter algorithm described in
    RFC4647.

    `languages` must be a sequence of `LanguageTag` instances which are
    to be filtered.

    `ranges` must be an iterable which represent the basic language ranges to
    filter with, in priority order. The language ranges must be given as
    `LanguageRange` objects.

    Return an iterator of languages which matched any of the `ranges`. The
    sequence produced by the iterator is in match order and duplicate-free. The
    first range to match a language yields the language into the iterator, no
    other range can yield that language afterwards.
    """

    if LanguageRange.WILDCARD in ranges:
        yield from languages
        return

    found: set[LanguageTag] = set()

    for language_range in ranges:
        range_str = language_range.match_str
        for language in languages:
            if language in found:
                continue

            match_str = language.match_str
            if match_str == range_str:
                yield language
                found.add(language)
                continue

            if len(range_str) < len(match_str):
                if (
                    match_str[: len(range_str)] == range_str
                    and match_str[len(range_str)] == "-"
                ):
                    yield language
                    found.add(language)
                    continue


def lookup_language(
    languages: Sequence[LanguageTag], ranges: Iterable[LanguageRange]
) -> LanguageTag | None:
    """
    Look up a single language in the sequence `languages` using the lookup
    mechanism described in RFC4647. If no match is found, `None` is
    returned. Otherwise, the first matching language is returned.

    `languages` must be a sequence of `LanguageTag` objects, while
    `ranges` must be an iterable of `LanguageRange` objects.
    """

    for language_range in ranges:
        while True:
            try:
                return next(iter(basic_filter_languages(languages, [language_range])))
            except StopIteration:
                pass

            try:
                language_range = language_range.strip_rightmost()
            except ValueError:
                break

    return None


class LanguageMap(dict[LanguageTag | None, Any]):
    """
    A `dict` subclass specialized for holding `LanugageTag`
    instances as keys.

    In addition to the interface provided by `dict`, instances of this
    class also have the following methods:
    """

    def lookup(self, language_ranges: Sequence[LanguageRange]) -> Any:
        """
        Perform an RFC4647 language range lookup on the keys in the
        dictionary. `language_ranges` must be a sequence of
        `LanguageRange` instances.

        Return the entry in the dictionary with a key as produced by
        `lookup_language`. If `lookup_language` does not find a match and the
        mapping contains an entry with key `None`, that entry is
        returned, otherwise `KeyError` is raised.
        """
        keys = list(self.keys())
        try:
            keys.remove(None)
        except ValueError:
            pass

        keys = cast(list[LanguageTag], keys)
        keys.sort()
        key = lookup_language(keys, language_ranges)
        return self[key]

    def any(self) -> Any:
        """
        Returns any element from the language map, preferring the `None`
        key if it is available.

        Guarantees to always return the same element for a map with the same
        keys, even if the keys are iterated over in a different order.
        """
        if not self:
            raise ValueError("any() on empty map")

        try:
            return self[None]
        except KeyError:
            return self[min(self)]  # type: ignore



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import OOBData
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class OOB(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_oob,
                ns=Namespace.X_OOB,
                priority=15,
            ),
        ]

    def _process_message_oob(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        oob = stanza.getTag("x", namespace=Namespace.X_OOB)
        if oob is None:
            return

        url = oob.getTagData("url")
        if url is None:
            self._log.warning("OOB data without url")
            self._log.warning(stanza)
            return

        desc = oob.getTagData("desc")
        properties.oob = OOBData(url, desc)



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class PubsubTest(StanzaHandlerTest):

    def test_purge_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node="princely_musings",
                id=None,
                item=None,
                data=None,
                deleted=False,
                retracted=False,
                purged=True,
            )

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = """
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <purge node='princely_musings'/>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message",
                callback=_on_message,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            )
        )

        self.dispatcher.process_data(event)

    def test_delete_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node="princely_musings",
                id=None,
                item=None,
                data=None,
                deleted=True,
                retracted=False,
                purged=False,
            )

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = """
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <delete node='princely_musings'>
                    <redirect uri='xmpp:hamlet@denmark.lit?;node=blog'/>
                </delete>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message",
                callback=_on_message,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            )
        )

        self.dispatcher.process_data(event)

    def test_retracted_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node="princely_musings",
                id="ae890ac52d0df67ed7cfdf51b644e901",
                item=None,
                data=None,
                deleted=False,
                retracted=True,
                purged=False,
            )

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = """
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <items node='princely_musings'>
                    <retract id='ae890ac52d0df67ed7cfdf51b644e901'/>
                </items>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message",
                callback=_on_message,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import hashlib
from dataclasses import dataclass

from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.simplexml import Node
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.util import from_xs_boolean

if TYPE_CHECKING:
    from nbxmpp.client import Client


class SecurityLabels(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_security_label,
                ns=Namespace.SECLABEL,
                priority=15,
            ),
        ]

    def _process_message_security_label(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        security = stanza.getTag("securitylabel", namespace=Namespace.SECLABEL)
        if security is None:
            return

        try:
            security_label = SecurityLabel.from_node(security)
        except ValueError as error:
            self._log.warning(error)
            return

        properties.security_label = security_label

    @iq_request_task
    def request_catalog(self, jid: str):
        _task = yield

        response = yield _make_catalog_request(self._client.domain, jid)
        if response.isError():
            raise StanzaError(response)

        catalog_node = response.getTag("catalog", namespace=Namespace.SECLABEL_CATALOG)

        try:
            restrict = from_xs_boolean(catalog_node.getAttr("restrict"))
        except Exception:
            restrict = False

        items = catalog_node.getTags("item")

        labels: dict[str, SecurityLabel | None] = {}
        default: str | None = None
        for item in items:
            selector = item.getAttr("selector")
            if selector is None:
                continue

            if item.getAttr("default") == "true":
                default = selector

            security = item.getTag("securitylabel", namespace=Namespace.SECLABEL)
            if security is None:
                labels[selector] = None
                continue

            try:
                security_label = SecurityLabel.from_node(security)
            except ValueError:
                continue

            labels[selector] = security_label

        yield Catalog(labels=labels, default=default, restrict=restrict)


def _make_catalog_request(domain: str | None, jid: str) -> Iq:
    iq = Iq(typ="get", to=domain)
    iq.addChild(name="catalog", namespace=Namespace.SECLABEL_CATALOG, attrs={"to": jid})
    return iq


@dataclass
class Displaymarking:
    name: str
    fgcolor: str
    bgcolor: str

    def to_node(self) -> Node:
        displaymarking = Node(tag="displaymarking")
        if self.fgcolor and self.fgcolor != "#000":
            displaymarking.setAttr("fgcolor", self.fgcolor)

        if self.bgcolor and self.bgcolor != "#FFF":
            displaymarking.setAttr("bgcolor", self.bgcolor)

        if self.name:
            displaymarking.setData(self.name)

        return displaymarking

    @classmethod
    def from_node(cls, node: Node) -> Displaymarking:
        return cls(
            name=node.getData(),
            fgcolor=node.getAttr("fgcolor") or "#000",
            bgcolor=node.getAttr("bgcolor") or "#FFF",
        )


@dataclass
class SecurityLabel:
    displaymarking: Displaymarking | None
    label: Node

    def to_node(self) -> Node:
        security = Node(tag="securitylabel", attrs={"xmlns": Namespace.SECLABEL})
        if self.displaymarking is not None:
            security.addChild(node=self.displaymarking.to_node())
        security.addChild(node=self.label)
        return security

    @classmethod
    def from_node(cls, security: Node) -> SecurityLabel:
        displaymarking = security.getTag("displaymarking")
        if displaymarking is not None:
            displaymarking = Displaymarking.from_node(displaymarking)

        label = security.getTag("label")
        if label is None:
            raise ValueError("label node missing")

        return cls(displaymarking=displaymarking, label=label)

    def get_label_hash(self) -> str:
        sha = hashlib.sha512()
        sha.update(str(self.label).encode())
        return sha.hexdigest()


@dataclass
class Catalog:
    labels: dict[str, SecurityLabel | None]
    default: str
    restrict: bool

    def get_label_names(self) -> list[str]:
        return list(self.labels.keys())



import unittest

from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.util import error_factory


class TestErrorParsing(unittest.TestCase):

    def test_error_parsing(self):
        stanza = """
        <iq from='upload.montague.tld'
            id='step_03'
            to='romeo@montague.tld/garden'
            type='error'>
          <error type='modify'>
            <not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas' />
            <text xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'>File too large. The maximum file size is 20000 bytes</text>
            <text xml:lang='de' xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'>File zu groß. Erlaubt sind 20000 bytes</text>
            <file-too-large xmlns='urn:xmpp:http:upload:0'>
              <max-file-size>20000</max-file-size>
            </file-too-large>
          </error>
        </iq>"""

        error = error_factory(Iq(node=stanza))
        self.assertEqual(error.condition, "not-acceptable")
        self.assertEqual(error.app_condition, "file-too-large")
        self.assertEqual(
            error.get_text(), "File too large. The maximum file size is 20000 bytes"
        )
        self.assertEqual(error.get_text("de"), "File zu groß. Erlaubt sind 20000 bytes")
        self.assertEqual(error.type, "modify")
        self.assertEqual(error.id, "step_03")
        self.assertEqual(error.jid, JID.from_string("upload.montague.tld"))



# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# XEP-0425: Message Moderation
from __future__ import annotations

from typing import TYPE_CHECKING

import datetime as dt

from nbxmpp import JID
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.simplexml import Node
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import ModerationData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Moderation(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_moderation_0_message,
                typ="groupchat",
                ns=Namespace.FASTEN,
                priority=20,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_moderation_0_tombstone_message,
                typ="groupchat",
                ns=Namespace.MESSAGE_MODERATE,
                priority=20,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_moderation_1_message,
                typ="groupchat",
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
        ]

    @iq_request_task
    def send_moderation_request(
        self,
        namespace: str,
        muc_jid: JID,
        stanza_id: str,
        reason: str | None = None,
    ):
        _task = yield

        if namespace == Namespace.MESSAGE_MODERATE:
            response = yield _make_moderation_request_0(muc_jid, stanza_id, reason)
        else:
            response = yield _make_moderation_request_1(muc_jid, stanza_id, reason)

        yield process_response(response)

    def _process_moderation_1_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:

        retract, is_tombstone = _get_retract_element(stanza, properties)
        if retract is None:
            return

        stamp = _parse_moderation_timestamp(retract, is_tombstone, properties)

        moderated = retract.getTag("moderated", namespace=Namespace.MESSAGE_MODERATE_1)
        if moderated is None:
            return

        try:
            by = _parse_by_attr(moderated)
        except ValueError as error:
            self._log.warning("Unable to determine by attribute: %s", error)
            by = None

        occupant_id = moderated.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        if is_tombstone:
            stanza_id = properties.mam.id
        else:
            stanza_id = retract.getAttr("id")

        if stanza_id is None:
            self._log.warning("Unable to determine stanza-id")
            self._log.warning(stanza)
            raise NodeProcessed

        properties.moderation = ModerationData(
            stanza_id=stanza_id,
            by=by,
            reason=retract.getTagData("reason"),
            stamp=stamp,
            is_tombstone=is_tombstone,
            occupant_id=occupant_id,
        )

    def _process_moderation_0_tombstone_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:

        if not properties.is_mam_message:
            return

        moderated = stanza.getTag("moderated", namespace=Namespace.MESSAGE_MODERATE)
        if moderated is None:
            return

        properties.moderation = self._parse_moderated_0(
            moderated, properties, properties.mam.id
        )

    def _process_moderation_0_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:

        if not properties.jid.is_bare:
            return

        apply_to = stanza.getTag("apply-to", namespace=Namespace.FASTEN)
        if apply_to is None:
            return

        moderated = apply_to.getTag("moderated", namespace=Namespace.MESSAGE_MODERATE)
        if moderated is None:
            return

        stanza_id = apply_to.getAttr("id")
        if stanza_id is None:
            self._log.warning("apply-to element without stanza-id")
            self._log.warning(stanza)
            raise NodeProcessed

        properties.moderation = self._parse_moderated_0(
            moderated, properties, stanza_id
        )

    def _parse_moderated_0(
        self, moderated: Node, properties: MessageProperties, stanza_id: str
    ) -> ModerationData | None:

        retract, is_tombstone = _get_retract_element(moderated, properties)
        if retract is None:
            self._log.warning("Failed to find <retract/ed> element")
            return None

        try:
            by = _parse_by_attr(moderated)
        except ValueError as error:
            self._log.warning(error)
            by = None

        stamp = _parse_moderation_timestamp(retract, is_tombstone, properties)

        occupant_id = moderated.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        return ModerationData(
            stanza_id=stanza_id,
            by=by,
            occupant_id=occupant_id,
            reason=moderated.getTagData("reason"),
            stamp=stamp,
            is_tombstone=is_tombstone,
        )


def _get_retract_element(
    element: Node, properties: MessageProperties
) -> tuple[Node | None, bool]:
    """
    returns a tuple
        Node and a boolean value which signals if it is a tombstone
    """

    retract = element.getTag("retract", namespace=Namespace.MESSAGE_RETRACT)
    if retract is not None:
        return retract, False

    retract = element.getTag("retract", namespace=Namespace.MESSAGE_RETRACT_1)
    if retract is not None:
        return retract, False

    if not properties.is_mam_message:
        return None, False

    retracted = element.getTag("retracted", namespace=Namespace.MESSAGE_RETRACT)
    if retracted is not None:
        return retracted, True

    retracted = element.getTag("retracted", namespace=Namespace.MESSAGE_RETRACT_1)
    if retracted is not None:
        return retracted, True

    return None, False


def _parse_by_attr(moderated: Node) -> JID | None:
    by_attr = moderated.getAttr("by")
    if by_attr is None:
        return None

    try:
        return JID.from_string(by_attr)
    except Exception as error:
        raise ValueError("Invalid JID: %s, %s" % (by_attr, error))


def _parse_moderation_timestamp(
    retract: Node, is_tombstone: bool, properties: MessageProperties
) -> dt.datetime:

    if is_tombstone:
        stamp_attr = retract.getAttr("stamp")
        stamp = parse_datetime(stamp_attr, check_utc=True, convert="utc")
        if stamp is not None:
            assert isinstance(stamp, dt.datetime)
            return stamp

    if properties.is_mam_message:
        stamp = properties.mam.timestamp
    else:
        stamp = properties.timestamp

    return dt.datetime.fromtimestamp(stamp, dt.timezone.utc)


def _make_moderation_request_0(muc_jid: JID, stanza_id: str, reason: str | None) -> Iq:
    iq = Iq("set", Namespace.FASTEN, to=muc_jid)
    query = iq.setQuery(name="apply-to")
    query.setAttr("id", stanza_id)
    moderate = query.addChild(name="moderate", namespace=Namespace.MESSAGE_MODERATE)
    moderate.addChild(name="retract", namespace=Namespace.MESSAGE_RETRACT)
    if reason is not None:
        moderate.addChild(name="reason", payload=[reason])
    return iq


def _make_moderation_request_1(muc_jid: JID, stanza_id: str, reason: str | None) -> Iq:
    iq = Iq("set", Namespace.MESSAGE_MODERATE_1, to=muc_jid)
    query = iq.setQuery(name="moderate")
    query.setAttr("id", stanza_id)
    query.addChild(name="retract", namespace=Namespace.MESSAGE_RETRACT_1)
    if reason is not None:
        query.addChild(name="reason", payload=[reason])
    return iq



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Attention(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_attention,
                ns=Namespace.ATTENTION,
                priority=15,
            ),
        ]

    def _process_message_attention(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        attention = stanza.getTag("attention", namespace=Namespace.ATTENTION)
        if attention is None:
            return

        if properties.is_mam_message:
            return

        if properties.is_carbon_message and properties.carbon.is_sent:
            return

        if stanza.getTag("delay", namespace=Namespace.DELAY2) is not None:
            return

        properties.attention = True
